import { resolveDataBinds } from '../utils/common';
import { set as lodashSet } from 'lodash';

const DEFAULT_MAX_TIMEOUT = 10 * 1000;

export function getMetaInfoBySourceKey(sourceKey) {
  const [materialName, name] = sourceKey.split(':');
  return {
    materialName,
    name,
  };
}

export function emitEvent(
  trigger,
  listeners = [],
  args,
  scopeContext = {},
  context
) {
  const targetListeners = listeners.filter((l) => l.trigger === trigger);
  targetListeners.forEach(async (listener) => {
    // 当前非捕获Event，再判断冒泡行为
    if (!args?.event?.isCapturePhase && listener.noPropagation) {
      args?.customEventData?.stopPropagation();
    }

    // 判断捕获的执行，只有执行的捕获与配置的捕获一致时。才会执行。
    if (
      (listener?.isCapturePhase || false) ===
      (args?.customEventData?.isCapturePhase || false)
    ) {
      try {
        const res = await invokeListener(listener, args, scopeContext, context);
        const eventName = `${listener.key}.success`;
        const event = {
          detail:
            typeof res === 'object' && res !== null && !Array.isArray(res)
              ? {
                  /**
                   * @deprecated 历史bug
                   * value,origin,isCapturePhase
                   */
                  value: res,
                  origin: args.event,
                  isCapturePhase: !!args.event?.isCapturePhase,
                  ...res,
                }
              : res,
          origin: args.event,
          isCapturePhase: !!args.event?.isCapturePhase,
          name: eventName,
        };
        emitEvent(
          eventName,
          listeners,
          {
            ...args,
            event,
            customEventData: event,
          },
          scopeContext,
          context
        );
        if (res) {
          console.warn(
            `@deprecated 事件编排中后续事件(${eventName}) event.detail.value.xxx 已废弃，建议使用 event.detail 直接访问前一动作返回值`
          );
        }
      } catch (e) {
        const eventName = `${listener.key}.fail`;
        try {
          e.value = e;
          e.origin = args.event;
          e.isCapturePhase = !!args.event?.isCapturePhase;
        } catch (e) {}
        const event = {
          detail: e,
          origin: args.event,
          isCapturePhase: !!args.event?.isCapturePhase,
          name: eventName,
        };
        emitEvent(
          eventName,
          listeners,
          {
            ...args,
            event,
            customEventData: event,
          },
          scopeContext,
          context
        );
        console.warn(
          `@deprecated 事件编排中后续事件(${eventName}) event.detail.value.xxx 已废弃，建议使用 event.detail 直接访问前一动作返回值`
        );
        // 之前 invoke 内部catch 了错误，不会抛错
        // throw e
      }
    }
  });
}

async function invokeListener(
  { instanceFunction, data = {}, dataBinds = {} },
  args,
  scopeContext,
  context
) {
  // ToDo resolve databinds
  const action = instanceFunction;
  let maxTimeout = DEFAULT_MAX_TIMEOUT;
  // eslint-disable-next-line no-underscore-dangle
  if (data._maxTimeout === 'Infinity') maxTimeout = data._maxTimeout;
  const resolvedData = {
    ...data,
  };
  const resolvedDataBinds = resolveDataBinds(
    dataBinds,
    args.forItems,
    { event: args.event },
    scopeContext,
    true,
    context
  );

  // eslint-disable-next-line no-restricted-syntax
  for (const key in resolvedDataBinds) {
    if (
      resolvedDataBinds[key] &&
      resolvedDataBinds[key].__type === 'scopedValue'
    ) {
      try {
        lodashSet(
          resolvedData,
          key,
          resolvedDataBinds[key].getValue(scopeContext)
        );
      } catch (e) {
        lodashSet(resolvedData, key, '');
      }
    } else {
      lodashSet(resolvedData, key, resolvedDataBinds[key]);
    }
  }

  const params = {
    data: resolvedData,
    ...args,
  };

  try {
    if (maxTimeout === 'Infinity') {
      return await action(params);
    }
    const p = action(params);
    if (p instanceof Promise) {
      let timeout = null;
      const r = await Promise.race([
        new Promise((resolve, reject) => {
          timeout = setTimeout(() => {
            reject(new Error(`timeout in ${maxTimeout}ms`));
          }, maxTimeout);
        }),
        p,
      ]);
      if (timeout) {
        clearTimeout(timeout);
      }
      return r;
    }
    return p;
  } catch (e) {
    console.error('Action error: ', e);
    throw e;
  }
}
